@@ -5,6 +5,8 @@
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -55,29 +57,39 @@
 #include <stdio.h>
 #include <stdlib.h>
 
-#include "cache.h"
+#include "cache_varnishd.h"
 
+#include "cache/cache_objhead.h"
+#include "cache/cache_transport.h"
 
 #include "hash/hash_slinger.h"
+
 #include "vsha256.h"
-#include "vtim.h"
+
+struct rush {
+	unsigned		magic;
+#define RUSH_MAGIC		0xa1af5f01
+	VTAILQ_HEAD(,req)	reqs;
+};
 
 static const struct hash_slinger *hash;
 static struct objhead *private_oh;
 
+static void hsh_rush1(const struct worker *, struct objhead *,
+    struct rush *, int);
+static void hsh_rush2(struct worker *, struct rush *);
+static int hsh_deref_objhead(struct worker *wrk, struct objhead **poh);
+static int hsh_deref_objhead_unlock(struct worker *wrk, struct objhead **poh,
+    int);
+
 /*---------------------------------------------------------------------*/
 
-static struct objcore *
-hsh_NewObjCore(struct worker *wrk)
-{
-	struct objcore *oc;
+#define VCF_RETURN(x) const struct vcf_return VCF_##x[1] = { \
+	{ .name = #x, } \
+};
 
-	ALLOC_OBJ(oc, OBJCORE_MAGIC);
-	XXXAN(oc);
-	wrk->stats->n_objectcore++;
-	oc->flags |= OC_F_BUSY | OC_F_INCOMPLETE;
-	return (oc);
-}
+VCF_RETURNS()
+#undef VCF_RETURN
 
 /*---------------------------------------------------------------------*/
 
@@ -90,6 +102,7 @@ hsh_newobjhead(void)
 	XXXAN(oh);
 	oh->refcnt = 1;
 	VTAILQ_INIT(&oh->objcs);
+	VTAILQ_INIT(&oh->waitinglist);
 	Lck_New(&oh->mtx, lck_objhdr);
 	return (oh);
 }
@@ -99,12 +112,11 @@ hsh_newobjhead(void)
 static void
 hsh_prealloc(struct worker *wrk)
 {
-	struct waitinglist *wl;
 
 	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
 
 	if (wrk->nobjcore == NULL)
-		wrk->nobjcore = hsh_NewObjCore(wrk);
+		wrk->nobjcore = ObjNew(wrk);
 	CHECK_OBJ_NOTNULL(wrk->nobjcore, OBJCORE_MAGIC);
 
 	if (wrk->nobjhead == NULL) {
@@ -113,15 +125,6 @@ hsh_prealloc(struct worker *wrk)
 	}
 	CHECK_OBJ_NOTNULL(wrk->nobjhead, OBJHEAD_MAGIC);
 
-	if (wrk->nwaitinglist == NULL) {
-		ALLOC_OBJ(wl, WAITINGLIST_MAGIC);
-		XXXAN(wl);
-		VTAILQ_INIT(&wl->list);
-		wrk->nwaitinglist = wl;
-		wrk->stats->n_waitinglist++;
-	}
-	CHECK_OBJ_NOTNULL(wrk->nwaitinglist, WAITINGLIST_MAGIC);
-
 	if (hash->prep != NULL)
 		hash->prep(wrk);
 }
@@ -129,19 +132,19 @@ hsh_prealloc(struct worker *wrk)
 /*---------------------------------------------------------------------*/
 
 struct objcore *
-HSH_Private(struct worker *wrk)
+HSH_Private(const struct worker *wrk)
 {
 	struct objcore *oc;
 
 	CHECK_OBJ_NOTNULL(private_oh, OBJHEAD_MAGIC);
 
-	oc = hsh_NewObjCore(wrk);
+	oc = ObjNew(wrk);
 	AN(oc);
 	oc->refcnt = 1;
 	oc->objhead = private_oh;
-	oc->flags |= OC_F_PRIVATE | OC_F_BUSY;
+	oc->flags |= OC_F_PRIVATE;
 	Lck_Lock(&private_oh->mtx);
-	VTAILQ_INSERT_TAIL(&private_oh->objcs, oc, list);
+	VTAILQ_INSERT_TAIL(&private_oh->objcs, oc, hsh_list);
 	private_oh->refcnt++;
 	Lck_Unlock(&private_oh->mtx);
 	return (oc);
@@ -152,22 +155,15 @@ void
 HSH_Cleanup(struct worker *wrk)
 {
 
-	if (wrk->nobjcore != NULL) {
-		FREE_OBJ(wrk->nobjcore);
-		wrk->stats->n_objectcore--;
-		wrk->nobjcore = NULL;
-	}
+	if (wrk->nobjcore != NULL)
+		ObjDestroy(wrk, &wrk->nobjcore);
+
 	if (wrk->nobjhead != NULL) {
 		Lck_Delete(&wrk->nobjhead->mtx);
 		FREE_OBJ(wrk->nobjhead);
 		wrk->nobjhead = NULL;
 		wrk->stats->n_objecthead--;
 	}
-	if (wrk->nwaitinglist != NULL) {
-		FREE_OBJ(wrk->nwaitinglist);
-		wrk->nwaitinglist = NULL;
-		wrk->stats->n_waitinglist--;
-	}
 	if (wrk->nhashpriv != NULL) {
 		/* XXX: If needed, add slinger method for this */
 		free(wrk->nhashpriv);
@@ -176,11 +172,12 @@ HSH_Cleanup(struct worker *wrk)
 }
 
 void
-HSH_DeleteObjHead(struct worker *wrk, struct objhead *oh)
+HSH_DeleteObjHead(const struct worker *wrk, struct objhead *oh)
 {
 
 	AZ(oh->refcnt);
 	assert(VTAILQ_EMPTY(&oh->objcs));
+	assert(VTAILQ_EMPTY(&oh->waitinglist));
 	Lck_Delete(&oh->mtx);
 	wrk->stats->n_objecthead--;
 	FREE_OBJ(oh);
@@ -193,10 +190,10 @@ HSH_AddString(struct req *req, void *ctx, const char *str)
 	CHECK_OBJ_NOTNULL(req, REQ_MAGIC);
 	AN(ctx);
 	if (str != NULL) {
-		SHA256_Update(ctx, str, strlen(str));
+		VSHA256_Update(ctx, str, strlen(str));
 		VSLb(req->vsl, SLT_Hash, "%s", str);
 	} else
-		SHA256_Update(ctx, &str, 1);
+		VSHA256_Update(ctx, &str, 1);
 }
 
 /*---------------------------------------------------------------------
@@ -207,8 +204,8 @@ HSH_AddString(struct req *req, void *ctx, const char *str)
  */
 
 static struct hsh_magiclist {
-	unsigned char was[SHA256_LEN];
-	unsigned char now[SHA256_LEN];
+	unsigned char was[VSHA256_LEN];
+	unsigned char now[VSHA256_LEN];
 } hsh_magiclist[] = {
 	{ .now = {	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
 			0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
@@ -253,23 +250,23 @@ static struct hsh_magiclist {
 static void
 hsh_testmagic(void *result)
 {
-	int i, j;
-	static int nused = 0;
+	size_t i, j;
+	static size_t nused = 0;
 
 	for (i = 0; i < nused; i++)
-		if (!memcmp(hsh_magiclist[i].was, result, SHA256_LEN))
+		if (!memcmp(hsh_magiclist[i].was, result, VSHA256_LEN))
 			break;
 	if (i == nused && i < HSH_NMAGIC)
-		memcpy(hsh_magiclist[nused++].was, result, SHA256_LEN);
+		memcpy(hsh_magiclist[nused++].was, result, VSHA256_LEN);
 	if (i == nused)
 		return;
 	assert(i < HSH_NMAGIC);
 	fprintf(stderr, "HASHMAGIC: <");
-	for (j = 0; j < SHA256_LEN; j++)
+	for (j = 0; j < VSHA256_LEN; j++)
 		fprintf(stderr, "%02x", ((unsigned char*)result)[j]);
 	fprintf(stderr, "> -> <");
-	memcpy(result, hsh_magiclist[i].now, SHA256_LEN);
-	for (j = 0; j < SHA256_LEN; j++)
+	memcpy(result, hsh_magiclist[i].now, VSHA256_LEN);
+	for (j = 0; j < VSHA256_LEN; j++)
 		fprintf(stderr, "%02x", ((unsigned char*)result)[j]);
 	fprintf(stderr, ">\n");
 }
@@ -277,17 +274,24 @@ hsh_testmagic(void *result)
 /*---------------------------------------------------------------------
  * Insert an object which magically appears out of nowhere or, more likely,
  * comes off some persistent storage device.
- * Return it with a reference held.
+ * Insert it with a reference held.
  */
 
 void
-HSH_Insert(struct worker *wrk, const void *digest, struct objcore *oc)
+HSH_Insert(struct worker *wrk, const void *digest, struct objcore *oc,
+    struct ban *ban)
 {
 	struct objhead *oh;
+	struct rush rush;
 
 	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
 	AN(digest);
 	CHECK_OBJ_NOTNULL(oc, OBJCORE_MAGIC);
+	AN(ban);
+	AN(oc->flags & OC_F_BUSY);
+	AZ(oc->flags & OC_F_PRIVATE);
+	assert(oc->refcnt == 1);
+	INIT_OBJ(&rush, RUSH_MAGIC);
 
 	hsh_prealloc(wrk);
 
@@ -297,17 +301,27 @@ HSH_Insert(struct worker *wrk, const void *digest, struct objcore *oc)
 	Lck_AssertHeld(&oh->mtx);
 	assert(oh->refcnt > 0);
 
-	/* Insert (precreated) objcore in objecthead */
-	oc->refcnt = 1;
-	CHECK_OBJ_NOTNULL(oc, OBJCORE_MAGIC);
-	AZ(oc->flags & OC_F_BUSY);
-
-	VTAILQ_INSERT_HEAD(&oh->objcs, oc, list);
-	/* NB: do not deref objhead the new object inherits our reference */
+	/* Mark object busy and insert (precreated) objcore in
+	   objecthead. The new object inherits our objhead reference. */
 	oc->objhead = oh;
+	VTAILQ_INSERT_TAIL(&oh->objcs, oc, hsh_list);
+	oc->refcnt++;				// For EXP_Insert
 	Lck_Unlock(&oh->mtx);
-	wrk->stats->n_objectcore++;
-	wrk->stats->n_vampireobject++;
+
+	BAN_RefBan(oc, ban);
+	AN(oc->ban);
+	EXP_Insert(wrk, oc);
+
+	/* Move the object first in the oh list, unbusy it and run the
+	   waitinglist if necessary */
+	Lck_Lock(&oh->mtx);
+	VTAILQ_REMOVE(&oh->objcs, oc, hsh_list);
+	VTAILQ_INSERT_HEAD(&oh->objcs, oc, hsh_list);
+	oc->flags &= ~OC_F_BUSY;
+	if (!VTAILQ_EMPTY(&oh->waitinglist))
+		hsh_rush1(wrk, oh, &rush, HSH_RUSH_POLICY);
+	Lck_Unlock(&oh->mtx);
+	hsh_rush2(wrk, &rush);
 }
 
 /*---------------------------------------------------------------------
@@ -328,7 +342,7 @@ hsh_insert_busyobj(struct worker *wrk, struct objhead *oh)
 	AN(oc->flags & OC_F_BUSY);
 	oc->refcnt = 1;		/* Owned by busyobj */
 	oc->objhead = oh;
-	VTAILQ_INSERT_TAIL(&oh->objcs, oc, list);
+	VTAILQ_INSERT_TAIL(&oh->objcs, oc, hsh_list);
 	return (oc);
 }
 
@@ -336,17 +350,18 @@ hsh_insert_busyobj(struct worker *wrk, struct objhead *oh)
  */
 
 enum lookup_e
-HSH_Lookup(struct req *req, struct objcore **ocp, struct objcore **bocp,
-    int wait_for_busy, int always_insert)
+HSH_Lookup(struct req *req, struct objcore **ocp, struct objcore **bocp)
 {
 	struct worker *wrk;
 	struct objhead *oh;
 	struct objcore *oc;
 	struct objcore *exp_oc;
-	double exp_t_origin;
+	const struct vcf_return *vr;
+	vtim_real exp_t_origin;
 	int busy_found;
-	enum lookup_e retval;
-	uint8_t *vary;
+	const uint8_t *vary;
+	unsigned xid = 0;
+	float dttl = 0.0;
 
 	AN(ocp);
 	*ocp = NULL;
@@ -357,6 +372,7 @@ HSH_Lookup(struct req *req, struct objcore **ocp, struct objcore **bocp,
 	wrk = req->wrk;
 	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
 	CHECK_OBJ_NOTNULL(req->http, HTTP_MAGIC);
+	CHECK_OBJ_ORNULL(req->vcf, VCF_MAGIC);
 	AN(hash);
 
 	hsh_prealloc(wrk);
@@ -380,7 +396,7 @@ HSH_Lookup(struct req *req, struct objcore **ocp, struct objcore **bocp,
 	CHECK_OBJ_NOTNULL(oh, OBJHEAD_MAGIC);
 	Lck_AssertHeld(&oh->mtx);
 
-	if (always_insert) {
+	if (req->hash_always_miss) {
 		/* XXX: should we do predictive Vary in this case ? */
 		/* Insert new objcore in objecthead and release mutex */
 		*bocp = hsh_insert_busyobj(wrk, oh);
@@ -393,111 +409,152 @@ HSH_Lookup(struct req *req, struct objcore **ocp, struct objcore **bocp,
 	busy_found = 0;
 	exp_oc = NULL;
 	exp_t_origin = 0.0;
-	VTAILQ_FOREACH(oc, &oh->objcs, list) {
+	VTAILQ_FOREACH(oc, &oh->objcs, hsh_list) {
 		/* Must be at least our own ref + the objcore we examine */
 		assert(oh->refcnt > 1);
 		CHECK_OBJ_NOTNULL(oc, OBJCORE_MAGIC);
 		assert(oc->objhead == oh);
+		assert(oc->refcnt > 0);
 
-		if (oc->exp_flags & OC_EF_DYING)
+		if (oc->flags & OC_F_DYING)
 			continue;
 		if (oc->flags & OC_F_FAILED)
 			continue;
 
+		CHECK_OBJ_ORNULL(oc->boc, BOC_MAGIC);
 		if (oc->flags & OC_F_BUSY) {
-			CHECK_OBJ_ORNULL(oc->busyobj, BUSYOBJ_MAGIC);
 			if (req->hash_ignore_busy)
 				continue;
 
-			if (oc->busyobj != NULL &&
-			    oc->busyobj->vary != NULL &&
-			    !VRY_Match(req, oc->busyobj->vary))
+			if (oc->boc && oc->boc->vary != NULL &&
+			    !VRY_Match(req, oc->boc->vary))
 				continue;
 
 			busy_found = 1;
 			continue;
 		}
 
-		if (oc->exp.ttl <= 0.)
+		if (oc->ttl <= 0.)
 			continue;
 
-		if (BAN_CheckObject(wrk, oc, req))
+		if (BAN_CheckObject(wrk, oc, req)) {
+			oc->flags |= OC_F_DYING;
+			EXP_Remove(oc);
 			continue;
+		}
 
-		vary = ObjGetattr(wrk, oc, OA_VARY, NULL);
+		if (ObjHasAttr(wrk, oc, OA_VARY)) {
+			vary = ObjGetAttr(wrk, oc, OA_VARY, NULL);
+			AN(vary);
+			if (!VRY_Match(req, vary))
+				continue;
+		}
 
-		if (vary != NULL && !VRY_Match(req, vary))
-			continue;
+		if (req->vcf != NULL) {
+			vr = req->vcf->func(req, &oc, &exp_oc, 0);
+			if (vr == VCF_CONTINUE)
+				continue;
+			if (vr == VCF_MISS) {
+				oc = NULL;
+				break;
+			}
+			if (vr == VCF_HIT)
+				break;
+			assert(vr == VCF_DEFAULT);
+		}
 
-		if (EXP_Ttl(req, &oc->exp) >= req->t_req) {
-			/* If still valid, use it */
+		if (EXP_Ttl(req, oc) > req->t_req) {
 			assert(oh->refcnt > 1);
 			assert(oc->objhead == oh);
-			oc->refcnt++;
-			if (oc->hits < LONG_MAX)
-				oc->hits++;
-			Lck_Unlock(&oh->mtx);
-			assert(HSH_DerefObjHead(wrk, &oh));
-			*ocp = oc;
-			return (HSH_HIT);
+			break;
 		}
-		if (EXP_Ttl(NULL, &oc->exp) < req->t_req && /* ignore req.ttl */
-		    oc->exp.t_origin > exp_t_origin) {
+
+		if (EXP_Ttl(NULL, oc) < req->t_req && /* ignore req.ttl */
+		    oc->t_origin > exp_t_origin) {
 			/* record the newest object */
 			exp_oc = oc;
-			exp_t_origin = oc->exp.t_origin;
+			exp_t_origin = oc->t_origin;
+			assert(oh->refcnt > 1);
+			assert(exp_oc->objhead == oh);
 		}
 	}
 
-	if (exp_oc != NULL) {
-		assert(oh->refcnt > 1);
-		assert(exp_oc->objhead == oh);
-		exp_oc->refcnt++;
+	if (req->vcf != NULL)
+		(void)req->vcf->func(req, &oc, &exp_oc, 1);
 
-		if (!busy_found) {
+	if (oc != NULL && oc->flags & OC_F_HFP) {
+		xid = ObjGetXID(wrk, oc);
+		dttl = EXP_Dttl(req, oc);
+		AN(hsh_deref_objhead_unlock(wrk, &oh, HSH_RUSH_POLICY));
+		wrk->stats->cache_hitpass++;
+		VSLb(req->vsl, SLT_HitPass, "%u %.6f", xid, dttl);
+		return (HSH_HITPASS);
+	}
+
+	if (oc != NULL) {
+		*ocp = oc;
+		oc->refcnt++;
+		if (oc->flags & OC_F_HFM) {
+			xid = ObjGetXID(wrk, oc);
+			dttl = EXP_Dttl(req, oc);
 			*bocp = hsh_insert_busyobj(wrk, oh);
-			retval = HSH_EXPBUSY;
-		} else {
-			AZ(req->hash_ignore_busy);
-			retval = HSH_EXP;
+			Lck_Unlock(&oh->mtx);
+			wrk->stats->cache_hitmiss++;
+			VSLb(req->vsl, SLT_HitMiss, "%u %.6f", xid, dttl);
+			return (HSH_HITMISS);
 		}
-		if (exp_oc->hits < LONG_MAX)
-			exp_oc->hits++;
+		oc->hits++;
+		AN(hsh_deref_objhead_unlock(wrk, &oh, HSH_RUSH_POLICY));
+		return (HSH_HIT);
+	}
+
+	if (exp_oc != NULL && exp_oc->flags & OC_F_HFM) {
+		/*
+		 * expired HFM ("grace/keep HFM")
+		 *
+		 * XXX should HFM objects actually have grace/keep ?
+		 * XXX also:  why isn't *ocp = exp_oc ?
+		 */
+		xid = ObjGetXID(wrk, exp_oc);
+		dttl = EXP_Dttl(req, exp_oc);
+		*bocp = hsh_insert_busyobj(wrk, oh);
 		Lck_Unlock(&oh->mtx);
-		if (retval == HSH_EXP)
-			assert(HSH_DerefObjHead(wrk, &oh));
-		*ocp = exp_oc;
-		return (retval);
+		wrk->stats->cache_hitmiss++;
+		VSLb(req->vsl, SLT_HitMiss, "%u %.6f", xid, dttl);
+		return (HSH_HITMISS);
 	}
 
 	if (!busy_found) {
-		/* Insert objcore in objecthead and release mutex */
 		*bocp = hsh_insert_busyobj(wrk, oh);
-		/* NB: no deref of objhead, new object inherits reference */
+
+		if (exp_oc != NULL) {
+			exp_oc->refcnt++;
+			*ocp = exp_oc;
+			if (EXP_Ttl_grace(req, exp_oc) >= req->t_req) {
+				exp_oc->hits++;
+				Lck_Unlock(&oh->mtx);
+				return (HSH_GRACE);
+			}
+		}
 		Lck_Unlock(&oh->mtx);
 		return (HSH_MISS);
 	}
 
+	AN(busy_found);
+	if (exp_oc != NULL && EXP_Ttl_grace(req, exp_oc) >= req->t_req) {
+		/* we do not wait on the busy object if in grace */
+		exp_oc->refcnt++;
+		*ocp = exp_oc;
+		exp_oc->hits++;
+		AN(hsh_deref_objhead_unlock(wrk, &oh, 0));
+		return (HSH_GRACE);
+	}
+
 	/* There are one or more busy objects, wait for them */
+	VTAILQ_INSERT_TAIL(&oh->waitinglist, req, w_list);
 
 	AZ(req->hash_ignore_busy);
 
-	if (wait_for_busy) {
-		CHECK_OBJ_NOTNULL(wrk->nwaitinglist, WAITINGLIST_MAGIC);
-		if (oh->waitinglist == NULL) {
-			oh->waitinglist = wrk->nwaitinglist;
-			wrk->nwaitinglist = NULL;
-		}
-		VTAILQ_INSERT_TAIL(&oh->waitinglist->list,
-		    req, w_list);
-		if (DO_DEBUG(DBG_WAITINGLIST))
-			VSLb(req->vsl, SLT_Debug, "on waiting list <%p>", oh);
-	} else {
-		if (DO_DEBUG(DBG_WAITINGLIST))
-			VSLb(req->vsl, SLT_Debug, "hit busy obj <%p>", oh);
-	}
-
-	wrk->stats->busy_sleep++;
 	/*
 	 * The objhead reference transfers to the sess, we get it
 	 * back when the sess comes off the waiting list and
@@ -505,67 +562,80 @@ HSH_Lookup(struct req *req, struct objcore **ocp, struct objcore **bocp,
 	 */
 	req->hash_objhead = oh;
 	req->wrk = NULL;
+	req->waitinglist = 1;
+
+	if (DO_DEBUG(DBG_WAITINGLIST))
+		VSLb(req->vsl, SLT_Debug, "on waiting list <%p>", oh);
+
 	Lck_Unlock(&oh->mtx);
+
+	wrk->stats->busy_sleep++;
 	return (HSH_BUSY);
 }
 
 /*---------------------------------------------------------------------
+ * Pick the req's we are going to rush from the waiting list
  */
 
 static void
-hsh_rush(struct worker *wrk, struct objhead *oh)
+hsh_rush1(const struct worker *wrk, struct objhead *oh, struct rush *r, int max)
 {
-	unsigned u;
+	int i;
 	struct req *req;
-	struct sess *sp;
-	struct waitinglist *wl;
+
+	if (max == 0)
+		return;
+	if (max == HSH_RUSH_POLICY)
+		max = cache_param->rush_exponent;
+	assert(max > 0);
 
 	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
 	CHECK_OBJ_NOTNULL(oh, OBJHEAD_MAGIC);
+	CHECK_OBJ_NOTNULL(r, RUSH_MAGIC);
+	VTAILQ_INIT(&r->reqs);
 	Lck_AssertHeld(&oh->mtx);
-	wl = oh->waitinglist;
-	CHECK_OBJ_ORNULL(wl, WAITINGLIST_MAGIC);
-	if (wl == NULL)
-		return;
-	for (u = 0; u < cache_param->rush_exponent; u++) {
-		req = VTAILQ_FIRST(&wl->list);
+	for (i = 0; i < max; i++) {
+		req = VTAILQ_FIRST(&oh->waitinglist);
 		if (req == NULL)
 			break;
 		CHECK_OBJ_NOTNULL(req, REQ_MAGIC);
 		wrk->stats->busy_wakeup++;
 		AZ(req->wrk);
-		VTAILQ_REMOVE(&wl->list, req, w_list);
-		DSL(DBG_WAITINGLIST, req->vsl->wid, "off waiting list");
-		if (SES_Reschedule_Req(req)) {
-			/*
-			 * In case of overloads, we ditch the entire
-			 * waiting list.
-			 */
-			wrk->stats->busy_wakeup--;
-			while (1) {
-				wrk->stats->busy_killed++;
-				AN (req->vcl);
-				VCL_Rel(&req->vcl);
-				sp = req->sp;
-				CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
-				CNT_AcctLogCharge(wrk->stats, req);
-				Req_Release(req);
-				SES_Delete(sp, SC_OVERLOAD, NAN);
-				req = VTAILQ_FIRST(&wl->list);
-				if (req == NULL)
-					break;
-				CHECK_OBJ_NOTNULL(req, REQ_MAGIC);
-				VTAILQ_REMOVE(&wl->list, req, w_list);
-				DSL(DBG_WAITINGLIST, req->vsl->wid,
-				    "kill from waiting list");
-			}
-			break;
-		}
+		VTAILQ_REMOVE(&oh->waitinglist, req, w_list);
+		VTAILQ_INSERT_TAIL(&r->reqs, req, w_list);
+		req->waitinglist = 0;
 	}
-	if (VTAILQ_EMPTY(&wl->list)) {
-		oh->waitinglist = NULL;
-		FREE_OBJ(wl);
-		wrk->stats->n_waitinglist--;
+}
+
+/*---------------------------------------------------------------------
+ * Rush req's that came from waiting list.
+ */
+
+static void
+hsh_rush2(struct worker *wrk, struct rush *r)
+{
+	struct req *req;
+
+	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
+	CHECK_OBJ_NOTNULL(r, RUSH_MAGIC);
+
+	while (!VTAILQ_EMPTY(&r->reqs)) {
+		req = VTAILQ_FIRST(&r->reqs);
+		CHECK_OBJ_NOTNULL(req, REQ_MAGIC);
+		VTAILQ_REMOVE(&r->reqs, req, w_list);
+		DSL(DBG_WAITINGLIST, req->vsl->wid, "off waiting list");
+		if (req->transport->reembark != NULL) {
+			// For ESI includes
+			req->transport->reembark(wrk, req);
+		} else {
+			/*
+			 * We ignore the queue limits which apply to new
+			 * requests because if we fail to reschedule there
+			 * may be vmod_privs to cleanup and we need a proper
+			 * workerthread for that.
+			 */
+			AZ(Pool_Task(req->sp->pool, &req->task, TASK_QUEUE_RUSH));
+		}
 	}
 }
 
@@ -573,18 +643,17 @@ hsh_rush(struct worker *wrk, struct objhead *oh)
  * Purge an entire objhead
  */
 
-void
-HSH_Purge(struct worker *wrk, struct objhead *oh, double ttl, double grace,
-double keep)
+unsigned
+HSH_Purge(struct worker *wrk, struct objhead *oh, vtim_real ttl_now,
+    vtim_dur ttl, vtim_dur grace, vtim_dur keep)
 {
 	struct objcore *oc, **ocp;
 	unsigned spc, ospc, nobj, n, n_tot = 0;
 	int more = 0;
-	double now;
 
 	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
 	CHECK_OBJ_NOTNULL(oh, OBJHEAD_MAGIC);
-	ospc = WS_Reserve(wrk->aws, 0);
+	ospc = WS_ReserveAll(wrk->aws);
 	assert(ospc >= sizeof *ocp);
 	/*
 	 * Because of "soft" purges, there might be oc's in the list that has
@@ -593,7 +662,7 @@ double keep)
 	 */
 	Lck_Lock(&oh->mtx);
 	assert(oh->refcnt > 0);
-	VTAILQ_FOREACH(oc, &oh->objcs, list) {
+	VTAILQ_FOREACH(oc, &oh->objcs, hsh_list) {
 		CHECK_OBJ_NOTNULL(oc, OBJCORE_MAGIC);
 		assert(oc->objhead == oh);
 		oc->flags &= ~OC_F_PURGED;
@@ -607,8 +676,7 @@ double keep)
 		ocp = (void*)wrk->aws->f;
 		Lck_Lock(&oh->mtx);
 		assert(oh->refcnt > 0);
-		now = VTIM_real();
-		VTAILQ_FOREACH(oc, &oh->objcs, list) {
+		VTAILQ_FOREACH(oc, &oh->objcs, hsh_list) {
 			CHECK_OBJ_NOTNULL(oc, OBJCORE_MAGIC);
 			assert(oc->objhead == oh);
 			if (oc->flags & OC_F_BUSY) {
@@ -621,7 +689,7 @@ double keep)
 				 */
 				continue;
 			}
-			if (oc->exp_flags & OC_EF_DYING)
+			if (oc->flags & OC_F_DYING)
 				continue;
 			if (oc->flags & OC_F_PURGED) {
 				/*
@@ -647,13 +715,14 @@ double keep)
 		for (n = 0; n < nobj; n++) {
 			oc = ocp[n];
 			CHECK_OBJ_NOTNULL(oc, OBJCORE_MAGIC);
-			EXP_Rearm(oc, now, ttl, grace, keep);
-			(void)HSH_DerefObjCore(wrk, &oc);
+			EXP_Rearm(oc, ttl_now, ttl, grace, keep);
+			(void)HSH_DerefObjCore(wrk, &oc, 0);
 		}
 		n_tot += nobj;
 	} while (more);
 	WS_Release(wrk->aws, 0);
 	Pool_PurgeStat(n_tot);
+	return (n_tot);
 }
 
 /*---------------------------------------------------------------------
@@ -678,17 +747,15 @@ HSH_Fail(struct objcore *oc)
 
 	Lck_Lock(&oh->mtx);
 	oc->flags |= OC_F_FAILED;
-	oc->flags &= ~OC_F_INCOMPLETE;
-	oc->busyobj = NULL;
 	Lck_Unlock(&oh->mtx);
 }
 
 /*---------------------------------------------------------------------
- * Remove the busyobj from an objcore
+ * Abandon a fetch we will not need
  */
 
-void
-HSH_Complete(struct objcore *oc)
+static void
+hsh_abandon(struct objcore *oc)
 {
 	struct objhead *oh;
 
@@ -697,11 +764,47 @@ HSH_Complete(struct objcore *oc)
 	CHECK_OBJ(oh, OBJHEAD_MAGIC);
 
 	Lck_Lock(&oh->mtx);
-	oc->busyobj = NULL;
-	oc->flags &= ~OC_F_INCOMPLETE;
+	oc->flags |= OC_F_ABANDON;
 	Lck_Unlock(&oh->mtx);
 }
 
+/*---------------------------------------------------------------------
+ * Cancel a fetch when the client does not need it any more
+ */
+
+void
+HSH_Cancel(struct worker *wrk, struct objcore *oc, struct boc *boc)
+{
+	struct boc *bocref = NULL;
+
+	CHECK_OBJ_NOTNULL(oc, OBJCORE_MAGIC);
+
+	if ((oc->flags & (OC_F_PRIVATE | OC_F_HFM | OC_F_HFP)) == 0)
+		return;
+
+	/*
+	 * NB: we use two distinct variables to only release the reference if
+	 * we had to acquire one. The caller-provided boc is optional.
+	 */
+	if (boc == NULL)
+		bocref = boc = HSH_RefBoc(oc);
+
+	CHECK_OBJ_ORNULL(boc, BOC_MAGIC);
+
+	if (oc->flags & OC_F_HFP)
+		AN(oc->flags & OC_F_HFM);
+
+	if (boc != NULL) {
+		hsh_abandon(oc);
+		ObjWaitState(oc, BOS_FINISHED);
+	}
+
+	if (bocref != NULL)
+		HSH_DerefBoc(wrk, oc);
+
+	ObjSlim(wrk, oc);
+}
+
 /*---------------------------------------------------------------------
  * Unbusy an objcore when the object is completely fetched.
  */
@@ -710,35 +813,92 @@ void
 HSH_Unbusy(struct worker *wrk, struct objcore *oc)
 {
 	struct objhead *oh;
+	struct rush rush;
 
 	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
 	CHECK_OBJ_NOTNULL(oc, OBJCORE_MAGIC);
 	oh = oc->objhead;
 	CHECK_OBJ(oh, OBJHEAD_MAGIC);
+	INIT_OBJ(&rush, RUSH_MAGIC);
 
 	AN(oc->stobj->stevedore);
 	AN(oc->flags & OC_F_BUSY);
 	assert(oh->refcnt > 0);
+	assert(oc->refcnt > 0);
 
 	if (!(oc->flags & OC_F_PRIVATE)) {
 		BAN_NewObjCore(oc);
-		EXP_Insert(wrk, oc);
-		AN(oc->exp_flags & OC_EF_EXP);
 		AN(oc->ban);
 	}
 
 	/* XXX: pretouch neighbors on oh->objcs to prevent page-on under mtx */
 	Lck_Lock(&oh->mtx);
 	assert(oh->refcnt > 0);
+	assert(oc->refcnt > 0);
+	if (!(oc->flags & OC_F_PRIVATE))
+		oc->refcnt++;			// For EXP_Insert
 	/* XXX: strictly speaking, we should sort in Date: order. */
-	VTAILQ_REMOVE(&oh->objcs, oc, list);
-	VTAILQ_INSERT_HEAD(&oh->objcs, oc, list);
+	VTAILQ_REMOVE(&oh->objcs, oc, hsh_list);
+	VTAILQ_INSERT_HEAD(&oh->objcs, oc, hsh_list);
 	oc->flags &= ~OC_F_BUSY;
-	if (oh->waitinglist != NULL)
-		hsh_rush(wrk, oh);
+	if (!VTAILQ_EMPTY(&oh->waitinglist)) {
+		assert(oh->refcnt > 1);
+		hsh_rush1(wrk, oh, &rush, HSH_RUSH_POLICY);
+	}
 	Lck_Unlock(&oh->mtx);
+	if (!(oc->flags & OC_F_PRIVATE))
+		EXP_Insert(wrk, oc);
+	hsh_rush2(wrk, &rush);
 }
 
+/*====================================================================
+ * HSH_Kill()
+ *
+ * It's dead Jim, kick it...
+ */
+
+void
+HSH_Kill(struct objcore *oc)
+{
+
+	CHECK_OBJ_NOTNULL(oc, OBJCORE_MAGIC);
+	CHECK_OBJ_NOTNULL(oc->objhead, OBJHEAD_MAGIC);
+
+	Lck_Lock(&oc->objhead->mtx);
+	oc->flags |= OC_F_DYING;
+	Lck_Unlock(&oc->objhead->mtx);
+	EXP_Remove(oc);
+}
+
+/*====================================================================
+ * HSH_Snipe()
+ *
+ * If objcore is idle, gain a ref and mark it dead.
+ */
+
+int
+HSH_Snipe(const struct worker *wrk, struct objcore *oc)
+{
+	int retval = 0;
+
+	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
+	CHECK_OBJ_NOTNULL(oc, OBJCORE_MAGIC);
+	CHECK_OBJ_NOTNULL(oc->objhead, OBJHEAD_MAGIC);
+
+	if (oc->refcnt == 1 && !Lck_Trylock(&oc->objhead->mtx)) {
+		if (oc->refcnt == 1 && !(oc->flags & OC_F_DYING)) {
+			oc->flags |= OC_F_DYING;
+			oc->refcnt++;
+			retval = 1;
+		}
+		Lck_Unlock(&oc->objhead->mtx);
+	}
+	if (retval)
+		EXP_Remove(oc);
+	return (retval);
+}
+
+
 /*---------------------------------------------------------------------
  * Gain a reference on an objcore
  */
@@ -761,23 +921,51 @@ HSH_Ref(struct objcore *oc)
  * Gain a reference on the busyobj, if the objcore has one
  */
 
-struct busyobj *
-HSH_RefBusy(const struct objcore *oc)
+struct boc *
+HSH_RefBoc(const struct objcore *oc)
 {
 	struct objhead *oh;
-	struct busyobj *bo;
+	struct boc *boc;
 
 	CHECK_OBJ_NOTNULL(oc, OBJCORE_MAGIC);
 	oh = oc->objhead;
 	CHECK_OBJ_NOTNULL(oh, OBJHEAD_MAGIC);
+	if (oc->boc == NULL)
+		return (NULL);
 	Lck_Lock(&oh->mtx);
 	assert(oc->refcnt > 0);
-	bo = oc->busyobj;
-	CHECK_OBJ_ORNULL(bo, BUSYOBJ_MAGIC);
-	if (bo != NULL)
-		bo->refcount++;
+	boc = oc->boc;
+	CHECK_OBJ_ORNULL(boc, BOC_MAGIC);
+	if (boc != NULL) {
+		assert(boc->refcount > 0);
+		if (boc->state < BOS_FINISHED)
+			boc->refcount++;
+		else
+			boc = NULL;
+	}
 	Lck_Unlock(&oh->mtx);
-	return (bo);
+	return (boc);
+}
+
+void
+HSH_DerefBoc(struct worker *wrk, struct objcore *oc)
+{
+	struct boc *boc;
+	unsigned r;
+
+	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
+	CHECK_OBJ_NOTNULL(oc, OBJCORE_MAGIC);
+	boc = oc->boc;
+	CHECK_OBJ_NOTNULL(boc, BOC_MAGIC);
+	Lck_Lock(&oc->objhead->mtx);
+	assert(oc->refcnt > 0);
+	assert(boc->refcount > 0);
+	r = --boc->refcount;
+	if (r == 0)
+		oc->boc = NULL;
+	Lck_Unlock(&oc->objhead->mtx);
+	if (r == 0)
+		ObjBocDone(wrk, oc, &boc);
 }
 
 /*--------------------------------------------------------------------
@@ -787,19 +975,17 @@ HSH_RefBusy(const struct objcore *oc)
  */
 
 int
-HSH_DerefObjCore(struct worker *wrk, struct objcore **ocp)
+HSH_DerefObjCore(struct worker *wrk, struct objcore **ocp, int rushmax)
 {
 	struct objcore *oc;
 	struct objhead *oh;
+	struct rush rush;
 	unsigned r;
 
-	AN(ocp);
-	oc = *ocp;
-	*ocp = NULL;
-
 	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
-	CHECK_OBJ_NOTNULL(oc, OBJCORE_MAGIC);
+	TAKE_OBJ_NOTNULL(oc, ocp, OBJCORE_MAGIC);
 	assert(oc->refcnt > 0);
+	INIT_OBJ(&rush, RUSH_MAGIC);
 
 	oh = oc->objhead;
 	CHECK_OBJ_NOTNULL(oh, OBJHEAD_MAGIC);
@@ -808,79 +994,83 @@ HSH_DerefObjCore(struct worker *wrk, struct objcore **ocp)
 	assert(oh->refcnt > 0);
 	r = --oc->refcnt;
 	if (!r)
-		VTAILQ_REMOVE(&oh->objcs, oc, list);
-	if (oh->waitinglist != NULL)
-		hsh_rush(wrk, oh);
+		VTAILQ_REMOVE(&oh->objcs, oc, hsh_list);
+	if (!VTAILQ_EMPTY(&oh->waitinglist)) {
+		assert(oh->refcnt > 1);
+		hsh_rush1(wrk, oh, &rush, rushmax);
+	}
 	Lck_Unlock(&oh->mtx);
+	hsh_rush2(wrk, &rush);
 	if (r != 0)
 		return (r);
 
+	AZ(oc->exp_flags);
+
 	BAN_DestroyObj(oc);
 	AZ(oc->ban);
 
 	if (oc->stobj->stevedore != NULL)
 		ObjFreeObj(wrk, oc);
-	FREE_OBJ(oc);
+	ObjDestroy(wrk, &oc);
 
-	wrk->stats->n_objectcore--;
 	/* Drop our ref on the objhead */
 	assert(oh->refcnt > 0);
-	(void)HSH_DerefObjHead(wrk, &oh);
+	(void)hsh_deref_objhead(wrk, &oh);
 	return (0);
 }
 
-int
-HSH_DerefObjHead(struct worker *wrk, struct objhead **poh)
+static int
+hsh_deref_objhead_unlock(struct worker *wrk, struct objhead **poh, int max)
 {
 	struct objhead *oh;
+	struct rush rush;
 	int r;
 
 	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
-	AN(poh);
-	oh = *poh;
-	*poh = NULL;
-	CHECK_OBJ_NOTNULL(oh, OBJHEAD_MAGIC);
+	TAKE_OBJ_NOTNULL(oh, poh, OBJHEAD_MAGIC);
+
+	Lck_AssertHeld(&oh->mtx);
 
 	if (oh == private_oh) {
-		AZ(oh->waitinglist);
-		Lck_Lock(&oh->mtx);
+		assert(VTAILQ_EMPTY(&oh->waitinglist));
 		assert(oh->refcnt > 1);
 		oh->refcnt--;
 		Lck_Unlock(&oh->mtx);
-		return(1);
+		return (1);
 	}
 
-	/*
-	 * Make absolutely certain that we do not let the final ref
-	 * disappear until the waitinglist is empty.
-	 * This is necessary because the req's on the waiting list do
-	 * not hold any ref on the objhead of their own, and we cannot
-	 * just make the hold the same ref's as objcore, that would
-	 * confuse hashers.
-	 */
-	while (oh->waitinglist != NULL) {
-		Lck_Lock(&oh->mtx);
-		assert(oh->refcnt > 0);
-		r = oh->refcnt;
-		hsh_rush(wrk, oh);
-		Lck_Unlock(&oh->mtx);
-		if (r > 1)
-			break;
-		usleep(100000);
+	INIT_OBJ(&rush, RUSH_MAGIC);
+	if (!VTAILQ_EMPTY(&oh->waitinglist)) {
+		assert(oh->refcnt > 1);
+		hsh_rush1(wrk, oh, &rush, max);
 	}
 
+	if (oh->refcnt == 1)
+		assert(VTAILQ_EMPTY(&oh->waitinglist));
+
 	assert(oh->refcnt > 0);
-	r = hash->deref(oh);
-	if (!r)
-		HSH_DeleteObjHead(wrk, oh);
+	r = hash->deref(wrk, oh); /* Unlocks oh->mtx */
+	hsh_rush2(wrk, &rush);
 	return (r);
 }
 
+static int
+hsh_deref_objhead(struct worker *wrk, struct objhead **poh)
+{
+	struct objhead *oh;
+
+	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
+	TAKE_OBJ_NOTNULL(oh, poh, OBJHEAD_MAGIC);
+
+	Lck_Lock(&oh->mtx);
+	return (hsh_deref_objhead_unlock(wrk, &oh, 0));
+}
+
 void
 HSH_Init(const struct hash_slinger *slinger)
 {
 
-	assert(DIGEST_LEN == SHA256_LEN);	/* avoid #include pollution */
+	assert(DIGEST_LEN == VSHA256_LEN);	/* avoid #include pollution */
 	hash = slinger;
 	if (hash->start != NULL)
 		hash->start();
