@@ -5,6 +5,8 @@
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -49,22 +51,23 @@
 #include "miniobj.h"
 #include "vas.h"
 #include "vqueue.h"
+#include "vtree.h"
 
 #include "vapi/vsl_int.h"
 
 /*--------------------------------------------------------------------*/
 
-enum body_status {
-#define BODYSTATUS(U,l)	BS_##U,
+struct body_status {
+	const char		*name;
+	int			nbr;
+	int			avail;
+	int			length_known;
+};
+
+#define BODYSTATUS(U, l, n, a, k) extern const struct body_status BS_##U[1];
 #include "tbl/body_status.h"
-};
 
-/*--------------------------------------------------------------------*/
-
-enum req_body_state_e {
-#define REQ_BODY(U)	REQ_BODY_##U,
-#include "tbl/req_body.h"
-};
+typedef const struct body_status *body_status_t;
 
 /*--------------------------------------------------------------------*/
 
@@ -86,6 +89,7 @@ enum {
 
 struct VSC_lck;
 struct VSC_main;
+struct VSC_main_wrk;
 struct backend;
 struct ban;
 struct ban_proto;
@@ -99,6 +103,7 @@ struct sess;
 struct transport;
 struct worker;
 struct listen_sock;
+struct vcf;
 
 #define DIGEST_LEN		32
 
@@ -163,7 +168,6 @@ struct http {
 	struct ws		*ws;
 	uint16_t		status;
 	uint8_t			protover;
-	uint8_t			conds;		/* If-* headers present */
 };
 
 /*--------------------------------------------------------------------*/
@@ -197,11 +201,7 @@ struct vxid_pool {
 
 /*--------------------------------------------------------------------*/
 
-struct vrt_privs {
-	unsigned		magic;
-#define VRT_PRIVS_MAGIC		0x03ba7501
-	VTAILQ_HEAD(,vrt_priv)	privs;
-};
+VRBT_HEAD(vrt_privs, vrt_priv);
 
 /* Worker pool stuff -------------------------------------------------*/
 
@@ -216,22 +216,20 @@ struct pool_task {
 /*
  * tasks are taken off the queues in this order
  *
- * prios up to TASK_QUEUE_RESERVE are run from the reserve
- *
  * TASK_QUEUE_{REQ|STR} are new req's (H1/H2), and subject to queue limit.
  *
- * TASK_QUEUE_RUSH is req's returning from waiting list, they are
- * not subject to TASK_QUEUE_CLIENT because we cannot safely clean
- * them up if scheduling them fails.
+ * TASK_QUEUE_RUSH is req's returning from waiting list
+ *
+ * NOTE: When changing the number of classes, update places marked with
+ * TASK_QUEUE__END in mgt_pool.c
  */
 enum task_prio {
 	TASK_QUEUE_BO,
-#define TASK_QUEUE_RESERVE	TASK_QUEUE_BO
 	TASK_QUEUE_RUSH,
 	TASK_QUEUE_REQ,
 	TASK_QUEUE_STR,
 	TASK_QUEUE_VCA,
-	TASK_QUEUE_END
+	TASK_QUEUE__END
 };
 
 #define TASK_QUEUE_CLIENT(prio) \
@@ -246,12 +244,12 @@ struct worker {
 	struct objhead		*nobjhead;
 	struct objcore		*nobjcore;
 	void			*nhashpriv;
-	struct VSC_main		*stats;
+	struct VSC_main_wrk	*stats;
 	struct vsl_log		*vsl;		// borrowed from req/bo
 
 	struct pool_task	task;
 
-	double			lastused;
+	vtim_real		lastused;
 	int			strangelove;
 
 	struct v1l		*v1l;
@@ -341,10 +339,11 @@ struct objcore {
 	struct storeobj		stobj[1];
 	struct objhead		*objhead;
 	struct boc		*boc;
-	double			timer_when;
-	long			hits;
+	vtim_real		timer_when;
+	VCL_INT			hits;
 
-	double			t_origin;
+
+	vtim_real		t_origin;
 	float			ttl;
 	float			grace;
 	float			keep;
@@ -356,7 +355,7 @@ struct objcore {
 	uint16_t		oa_present;
 
 	unsigned		timer_idx;	// XXX 4Gobj limit
-	double			last_lru;
+	vtim_real		last_lru;
 	VTAILQ_ENTRY(objcore)	hsh_list;
 	VTAILQ_ENTRY(objcore)	lru_list;
 	VTAILQ_ENTRY(objcore)	ban_list;
@@ -389,7 +388,7 @@ struct busyobj {
 	 * All fields from retries and down are zeroed when the busyobj
 	 * is recycled.
 	 */
-	int			retries;
+	unsigned		retries;
 	struct req		*req;
 	struct sess		*sp;
 	struct worker		*wrk;
@@ -402,9 +401,12 @@ struct busyobj {
 	struct http		*bereq0;
 	struct http		*bereq;
 	struct http		*beresp;
+	struct objcore		*bereq_body;
 	struct objcore		*stale_oc;
 	struct objcore		*fetch_objcore;
 
+	const char		*no_retry;
+
 	struct http_conn	*htc;
 
 	struct pool_task	fetch_task;
@@ -413,13 +415,13 @@ struct busyobj {
 #include "tbl/bo_flags.h"
 
 	/* Timeouts */
-	double			connect_timeout;
-	double			first_byte_timeout;
-	double			between_bytes_timeout;
+	vtim_dur		connect_timeout;
+	vtim_dur		first_byte_timeout;
+	vtim_dur		between_bytes_timeout;
 
 	/* Timers */
-	double			t_first;	/* First timestamp logged */
-	double			t_prev;		/* Previous timestamp logged */
+	vtim_real		t_first;	/* First timestamp logged */
+	vtim_real		t_prev;		/* Previous timestamp logged */
 
 	/* Acct */
 	struct acct_bereq	acct;
@@ -434,21 +436,32 @@ struct busyobj {
 
 	uint8_t			digest[DIGEST_LEN];
 	struct vrt_privs	privs[1];
+
+	uint16_t		err_code;
+	const char		*err_reason;
 };
 
 
 /*--------------------------------------------------------------------*/
 
+struct reqtop {
+	unsigned		magic;
+#define REQTOP_MAGIC		0x57fbda52
+	struct req		*topreq;
+	struct vcl		*vcl0;
+	struct vrt_privs	privs[1];
+};
+
 struct req {
 	unsigned		magic;
 #define REQ_MAGIC		0x2751aaa1
 
 	enum req_step		req_step;
-	volatile enum req_body_state_e	req_body_status;
+	body_status_t		req_body_status;
 	enum sess_close		doclose;
-	int			restarts;
-	int			esi_level;
-	struct req		*top;	/* esi_level == 0 request */
+	unsigned		restarts;
+	unsigned		esi_level;
+	struct reqtop		*top;	/* esi_level == 0 request */
 
 #define REQ_FLAG(l, r, w, d) unsigned	l:1;
 #include "tbl/req_flags.h"
@@ -477,10 +490,9 @@ struct req {
 
 	uint8_t			digest[DIGEST_LEN];
 
-	double			d_ttl;
-	double			d_grace;
+	vtim_dur		d_ttl;
+	vtim_dur		d_grace;
 
-	ssize_t			req_bodybytes;	/* Parsed req bodybytes */
 	const struct stevedore	*storage;
 
 	const struct director	*director_hint;
@@ -489,9 +501,9 @@ struct req {
 	uintptr_t		ws_req;		/* WS above request data */
 
 	/* Timestamps */
-	double			t_first;	/* First timestamp logged */
-	double			t_prev;		/* Previous timestamp logged */
-	double			t_req;		/* Headers complete */
+	vtim_real		t_first;	/* First timestamp logged */
+	vtim_real		t_prev;		/* Previous timestamp logged */
+	vtim_real		t_req;		/* Headers complete */
 
 	struct http_conn	*htc;
 	struct vfp_ctx		*vfc;
@@ -511,15 +523,11 @@ struct req {
 
 	/* Deliver pipeline */
 	struct vdp_ctx		*vdc;
+	const char		*filter_list;
 
 	/* Delivery mode */
 	unsigned		res_mode;
-#define RES_LEN			(1<<1)
-#define RES_EOF			(1<<2)
-#define RES_CHUNKED		(1<<3)
 #define RES_ESI			(1<<4)
-#define RES_ESI_CHILD		(1<<5)
-#define RES_GUNZIP		(1<<6)
 #define RES_PIPE		(1<<7)
 
 	/* Transaction VSL buffer */
@@ -529,8 +537,12 @@ struct req {
 	struct acct_req		acct;
 
 	struct vrt_privs	privs[1];
+
+	struct vcf		*vcf;
 };
 
+#define IS_TOPREQ(req) ((req)->top->topreq == (req))
+
 /*--------------------------------------------------------------------
  * Struct sess is a high memory-load structure because sessions typically
  * hang around the waiter for relatively long time.
@@ -543,7 +555,6 @@ struct req {
 
 enum sess_attr {
 #define SESS_ATTR(UP, low, typ, len)	SA_##UP,
-	SA_TRANSPORT,
 #include "tbl/sess_attr.h"
 	SA_LAST
 };
@@ -564,12 +575,17 @@ struct sess {
 
 	struct ws		ws[1];
 
-	/* Timestamps, all on TIM_real() timescale */
-	double			t_open;		/* fd accepted */
-	double			t_idle;		/* fd accepted or resp sent */
-
+	vtim_real		t_open;		/* fd accepted */
+	vtim_real		t_idle;		/* fd accepted or resp sent */
+	vtim_dur		timeout_idle;
+	vtim_dur		timeout_linger;
+	vtim_dur		send_timeout;
+	vtim_dur		idle_send_timeout;
 };
 
+#define SESS_TMO(sp, tmo)					\
+	(isnan((sp)->tmo) ? cache_param->tmo : (sp)->tmo)
+
 /* Prototypes etc ----------------------------------------------------*/
 
 
@@ -579,7 +595,7 @@ struct sess {
 struct ban_proto *BAN_Build(void);
 const char *BAN_AddTest(struct ban_proto *,
     const char *, const char *, const char *);
-const char *BAN_Commit(struct ban_proto *b, struct VSC_main *stats);
+const char *BAN_Commit(struct ban_proto *b);
 void BAN_Abandon(struct ban_proto *b);
 
 /* cache_cli.c [CLI] */
@@ -588,7 +604,8 @@ extern pthread_t cli_thread;
 
 /* cache_http.c */
 unsigned HTTP_estimate(unsigned nhttp);
-void HTTP_Copy(struct http *to, const struct http * const fm);
+void HTTP_Clone(struct http *to, const struct http * const fm);
+void HTTP_Dup(struct http *to, const struct http * const fm);
 struct http *HTTP_create(void *p, uint16_t nhttp, unsigned);
 const char *http_Status2Reason(unsigned, const char **);
 unsigned http_EstimateWS(const struct http *fm, unsigned how);
@@ -600,7 +617,7 @@ int HTTP_Decode(struct http *to, const uint8_t *fm);
 void http_ForceHeader(struct http *to, const char *hdr, const char *val);
 void http_PrintfHeader(struct http *to, const char *fmt, ...)
     v_printflike_(2, 3);
-void http_TimeHeader(struct http *to, const char *fmt, double now);
+void http_TimeHeader(struct http *to, const char *fmt, vtim_real now);
 void http_Proto(struct http *to);
 void http_SetHeader(struct http *to, const char *hdr);
 void http_SetH(struct http *to, unsigned n, const char *fm);
@@ -616,7 +633,7 @@ double http_GetHdrQ(const struct http *hp, const char *hdr, const char *field);
 ssize_t http_GetContentLength(const struct http *hp);
 uint16_t http_GetStatus(const struct http *hp);
 int http_IsStatus(const struct http *hp, int);
-void http_SetStatus(struct http *to, uint16_t status);
+void http_SetStatus(struct http *to, uint16_t status, const char *reason);
 const char *http_GetMethod(const struct http *hp);
 int http_HdrIs(const struct http *hp, const char *hdr, const char *val);
 void http_CopyHome(const struct http *hp);
@@ -662,7 +679,7 @@ int Lck__Owned(const struct lock *lck);
 
 /* public interface: */
 void Lck_Delete(struct lock *lck);
-int Lck_CondWait(pthread_cond_t *cond, struct lock *lck, double);
+int Lck_CondWait(pthread_cond_t *cond, struct lock *lck, vtim_real);
 
 #define Lck_New(a, b) Lck__New(a, b, #b)
 #define Lck_Lock(a) Lck__Lock(a, __func__, __LINE__)
@@ -686,7 +703,10 @@ int ObjHasAttr(struct worker *, struct objcore *, enum obj_attr);
 const void *ObjGetAttr(struct worker *, struct objcore *, enum obj_attr,
     ssize_t *len);
 
-typedef int objiterate_f(void *priv, int flush, const void *ptr, ssize_t len);
+typedef int objiterate_f(void *priv, unsigned flush,
+    const void *ptr, ssize_t len);
+#define OBJ_ITER_FLUSH	0x01
+#define OBJ_ITER_FINAL	0x02
 
 int ObjIterate(struct worker *, struct objcore *,
     void *priv, objiterate_f *func, int final);
@@ -708,16 +728,18 @@ const char *SES_Get_String_Attr(const struct sess *sp, enum sess_attr a);
 void VSLv(enum VSL_tag_e tag, uint32_t vxid, const char *fmt, va_list va);
 void VSL(enum VSL_tag_e tag, uint32_t vxid, const char *fmt, ...)
     v_printflike_(3, 4);
+void VSLs(enum VSL_tag_e tag, uint32_t vxid, const struct strands *s);
 void VSLbv(struct vsl_log *, enum VSL_tag_e tag, const char *fmt, va_list va);
 void VSLb(struct vsl_log *, enum VSL_tag_e tag, const char *fmt, ...)
     v_printflike_(3, 4);
 void VSLbt(struct vsl_log *, enum VSL_tag_e tag, txt t);
-void VSLb_ts(struct vsl_log *, const char *event, double first, double *pprev,
-    double now);
+void VSLbs(struct vsl_log *, enum VSL_tag_e tag, const struct strands *s);
+void VSLb_ts(struct vsl_log *, const char *event, vtim_real first,
+    vtim_real *pprev, vtim_real now);
 void VSLb_bin(struct vsl_log *, enum VSL_tag_e, ssize_t, const void*);
 
 static inline void
-VSLb_ts_req(struct req *req, const char *event, double now)
+VSLb_ts_req(struct req *req, const char *event, vtim_real now)
 {
 
 	if (isnan(req->t_first) || req->t_first == 0.)
@@ -726,7 +748,7 @@ VSLb_ts_req(struct req *req, const char *event, double now)
 }
 
 static inline void
-VSLb_ts_busyobj(struct busyobj *bo, const char *event, double now)
+VSLb_ts_busyobj(struct busyobj *bo, const char *event, vtim_real now)
 {
 
 	if (isnan(bo->t_first) || bo->t_first == 0.)
@@ -755,7 +777,11 @@ void WRK_BgThread(pthread_t *thr, const char *name, bgthread_t *func,
 /* cache_ws.c */
 
 void WS_Init(struct ws *ws, const char *id, void *space, unsigned len);
-unsigned WS_Reserve(struct ws *ws, unsigned bytes);
+
+/* WS_Reserve(): Use WS_ReserveSize() or WS_ReserveAll() */
+unsigned WS_Reserve(struct ws *ws, unsigned bytes) v_deprecated_;
+unsigned WS_ReserveSize(struct ws *, unsigned);
+unsigned WS_ReserveAll(struct ws *);
 unsigned WS_ReserveLumps(struct ws *ws, size_t sz);
 void WS_MarkOverflow(struct ws *ws);
 void WS_Release(struct ws *ws, unsigned bytes);
@@ -770,6 +796,9 @@ void *WS_Printf(struct ws *ws, const char *fmt, ...) v_printflike_(2, 3);
 int WS_Inside(const struct ws *, const void *, const void *);
 void WS_Assert_Allocated(const struct ws *ws, const void *ptr, ssize_t len);
 
+void WS_VSB_new(struct vsb *, struct ws *);
+char *WS_VSB_finish(struct vsb *, struct ws *, size_t *);
+
 static inline char*
 WS_Front(const struct ws *ws)
 {
@@ -777,13 +806,12 @@ WS_Front(const struct ws *ws)
 }
 
 /* cache_rfc2616.c */
-void RFC2616_Ttl(struct busyobj *, double now, double *t_origin,
+void RFC2616_Ttl(struct busyobj *, vtim_real now, vtim_real *t_origin,
     float *ttl, float *grace, float *keep);
 unsigned RFC2616_Req_Gzip(const struct http *);
 int RFC2616_Do_Cond(const struct req *sp);
 void RFC2616_Weaken_Etag(struct http *hp);
 void RFC2616_Vary_AE(struct http *hp);
-void RFC2616_Response_Body(const struct worker *, const struct busyobj *);
 
 /*
  * A normal pointer difference is signed, but we never want a negative value
