@@ -5,6 +5,8 @@
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -51,8 +53,14 @@
  * Whenever something is deleted or changed in a way which is not
  * binary/load-time compatible, increment MAJOR version
  *
- * unreleased (planned for 2020-03-15)
- *	(nothing yet)
+ * 11.0 (2020-03-16)
+ *	Changed type of vsa_suckaddr_len from int to size_t
+ *	New prefix_{ptr|len} fields in vrt_backend
+ *	VRT_HashStrands32() added
+ *	VRT_l_resp_body() changed
+ *	VRT_l_beresp_body() changed
+ *	VRT_Format_Proxy() added	// transitional interface
+ *	VRT_AllocStrandsWS() added
  * 10.0 (2019-09-15)
  *	VRT_UpperLowerStrands added.
  *	VRT_synth_page now takes STRANDS argument
@@ -85,6 +93,9 @@
  *	VRT_LookupDirector() added
  *	VRT_SetChanged() added
  *	VRT_SetHealth() removed
+ *	// in cache_filter.h:
+ *	VRT_AddVDP() added
+ *	VRT_RemoveVDP() added
  * 8.0 (2018-09-15)
  *	VRT_Strands() added
  *	VRT_StrandsWS() added
@@ -101,6 +112,9 @@
  *	VRT_SetHealth() added
  *	VRT_DisableDirector() added
  *	VRT_DelDirector() added
+ *	// in cache_filter.h:
+ *	VRT_AddVFP() added
+ *	VRT_RemoveVFP() added
  * 7.0 (2018-03-15)
  *	lots of stuff moved from cache.h to cache_varnishd.h
  *	   (ie: from "$Abi vrt" to "$Abi strict")
@@ -145,7 +159,7 @@
  *	vrt_acl type added
  */
 
-#define VRT_MAJOR_VERSION	10U
+#define VRT_MAJOR_VERSION	11U
 
 #define VRT_MINOR_VERSION	0U
 
@@ -172,17 +186,36 @@ struct vsl_log;
 struct ws;
 struct VSC_main;
 
+/*
+ * VCL_STRANDS:
+ *
+ * An argc+argv type of data structure where n indicates the number of strings
+ * in the p array. Individual components of a strands may be null.
+ *
+ * A STRANDS allows you to work on a strings concatenation with the option to
+ * collect it into a single STRING, or if possible work directly on individual
+ * parts.
+ *
+ * The memory management is very strict: a VMOD function receiving a STRANDS
+ * argument should keep no reference after the function returns. Retention of
+ * a STRANDS further in the ongoing task is undefined behavior and may result
+ * in a panic or data corruption.
+ */
+
 struct strands {
 	int		n;
 	const char	**p;
 };
 
+struct strands * VRT_AllocStrandsWS(struct ws *, int);
+
+
 /*
  * VCL_BLOB:
  *
- * opaque, immutable data (pointer + length), minimum lifetime is the VCL task.
+ * Opaque, immutable data (pointer + length), minimum lifetime is the VCL task.
  *
- * type (optional) is owned by the creator of the blob. blob consumers may use
+ * Type (optional) is owned by the creator of the blob. blob consumers may use
  * it for checks, but should not assert on it.
  *
  * The data behind the blob pointer is assumed to be immutable for the blob's
@@ -240,6 +273,11 @@ struct vrt_type {
 	size_t				szof;
 };
 
+enum lbody_e {
+	LBODY_SET,
+	LBODY_ADD,
+};
+
 /***********************************************************************
  * This is the composite argument we pass to compiled VCL and VRT
  * functions.
@@ -251,10 +289,15 @@ struct vrt_ctx {
 
 	unsigned			syntax;
 	unsigned			method;
-	unsigned			*handling;	// not in director context
+	unsigned			*handling;
 	unsigned			vclver;
 
-	struct vsb			*msg;	// Only in ...init()
+	/*
+	 * msg is for error messages and exists only for
+	 * VCL_EVENT_LOAD
+	 * VCL_EVENT_WARM
+	 */
+	struct vsb			*msg;
 	struct vsl_log			*vsl;
 	VCL_VCL				vcl;
 	struct ws			*ws;
@@ -338,7 +381,9 @@ extern const void * const vrt_magic_string_unset;
 	vtim_dur			first_byte_timeout;	\
 	vtim_dur			between_bytes_timeout;	\
 	unsigned			max_connections;	\
-	unsigned			proxy_header;
+	unsigned			proxy_header;		\
+	void				*prefix_ptr;		\
+	unsigned			prefix_len;
 
 #define VRT_BACKEND_HANDLE()			\
 	do {					\
@@ -357,7 +402,7 @@ extern const void * const vrt_magic_string_unset;
 
 struct vrt_backend {
 	unsigned			magic;
-#define VRT_BACKEND_MAGIC		0x4799ce6b
+#define VRT_BACKEND_MAGIC		0x4799ce6c
 	VRT_BACKEND_FIELDS(const)
 	VCL_IP				ipv4_suckaddr;
 	VCL_IP				ipv6_suckaddr;
@@ -516,6 +561,8 @@ void VRT_DelDirector(VCL_BACKEND *);
 
 /* Suckaddr related */
 int VRT_VSA_GetPtr(VRT_CTX, VCL_IP sua, const unsigned char ** dst);
+/* transitional interface */
+void VRT_Format_Proxy(struct vsb *, VCL_INT, VCL_IP, VCL_IP, VCL_STRING);
 
 typedef int vmod_event_f(VRT_CTX, struct vmod_priv *, enum vcl_event_e);
 
@@ -538,6 +585,7 @@ VCL_STEVEDORE VRT_stevedore(const char *nm);
 
 int VRT_CompareStrands(VCL_STRANDS a, VCL_STRANDS b);
 VCL_BOOL VRT_Strands2Bool(VCL_STRANDS);
+uint32_t VRT_HashStrands32(VCL_STRANDS);
 char *VRT_Strands(char *, size_t, VCL_STRANDS);
 VCL_STRING VRT_StrandsWS(struct ws *, const char *, VCL_STRANDS);
 VCL_STRING VRT_CollectStrands(VRT_CTX, VCL_STRANDS);
